--[[ ---------------------------------------------------------------------------

Acheron: death reports

Acheron is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Acheron is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with Acheron.	If not, see <http://www.gnu.org/licenses/>.

----------------------------------------------------------------------------- ]]

--[[ ---------------------------------------------------------------------------
	 Addon and libraries
----------------------------------------------------------------------------- ]]
Acheron = nil

if LibStub("LibLogger-1.0", true) then
    Acheron = LibStub("AceAddon-3.0"):NewAddon("Acheron", "AceEvent-3.0", "AceBucket-3.0", "AceTimer-3.0", "AceConsole-3.0", "AceHook-3.0", "LibLogger-1.0")
else
	Acheron = LibStub("AceAddon-3.0"):NewAddon("Acheron", "AceEvent-3.0", "AceBucket-3.0", "AceTimer-3.0", "AceConsole-3.0", "AceHook-3.0")
end

local L = LibStub("AceLocale-3.0"):GetLocale("Acheron")
local GUI = LibStub("AceGUI-3.0")
local QH = LibStub("LibQuickHealth-1.0", true)
local LDB = LibStub:GetLibrary("LibDataBroker-1.1")

local menuTypes= {"SELF", "PARTY", "RAID_PLAYER"}

local VER = " "..(GetAddOnMetadata("Acheron", "X-Curse-Packaged-Version") or GetAddOnMetadata("Acheron", "Version") or "")


--[[ ---------------------------------------------------------------------------
	 Ace3 initialization
----------------------------------------------------------------------------- ]]
function Acheron:OnInitialize()

	-- Initialize persistent addon variables
	self.combatLogs = {}
	self.combatLogTimeline = {first = 1, last = 0, logs = {}}
	self.guidNames = {}
	
	self.solo = nil
	self.updateUnitsBucket = nil
	self.currentAuraList = nil
	
	self.frame = nil
	self.frameReports = nil
	self.channels = nil
	self.availableReports = nil
	self.whisperEditBox = nil
	self.whisperTarget = nil
	
	-- Set up saved variables and config options
	self.db = LibStub("AceDB-3.0"):New("AcheronDB", self.defaults, "Default")

	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileDeleted","OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
	
	self.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Acheron", self.options, {"acheron"})
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Acheron", "Acheron")
	
	-- Logging
	if LibStub("LibLogger-1.0", true) then
		self:SetLogLevel(self.db.profile.logLevel)
		self:SetPerformanceMode(true)
	end
	
	-- Set up the data broker object
	self.dobj = LDB:NewDataObject("Acheron", {
		type = "launcher",
		icon = "Interface\\Icons\\Ability_Creature_Cursed_03",
    	OnClick = function(frame, button)
    		Acheron:ShowDeathReports()
        end,
		OnTooltipShow = function(tooltip)
			if tooltip and tooltip.AddLine then
				tooltip:AddLine("Acheron"..VER)
			end
		end
	})
	
	-- Confirmation dialogues
	StaticPopupDialogs["ACHERON_CLEAR_ALL"] = {
  		text = L["Clear Acheron?"],
  		button1 = L["Yes"],
  		button2 = L["No"],
  		OnAccept = function() Acheron:ClearReports() end,
  		timeout = 0,
  		whileDead = 1,
  		hideOnEscape = 1
	};
	
	StaticPopupDialogs["ACHERON_CLEAR"] = {
  		text = L["Clear Acheron for %s?"],
  		button1 = L["Yes"],
  		button2 = L["No"],
  		OnAccept = function(name) Acheron:ClearReports(name.data) end,
  		timeout = 0,
  		whileDead = 1,
  		hideOnEscape = 1
	};
	
	-- Create the display frame
	self:CreateFrame()
	self:DisplayWhiteList()
	self:DisplayBlackList()
	
	-- Register events
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "ZONE_CHANGED_NEW_AREA")
	
end


--[[ ---------------------------------------------------------------------------
	 Ace3 enable addon
----------------------------------------------------------------------------- ]]
function Acheron:OnEnable()

		self:DoEnable()

end


--[[ ---------------------------------------------------------------------------
	 Enables addon 
----------------------------------------------------------------------------- ]]
function Acheron:DoEnable()

	if not self:GetProfileParam("enable") then return end
	if self.debug then self:debug("Addon enabled") end
	
	-- Register events
	self.updateUnitsBucket = self:RegisterBucketEvent({"PARTY_MEMBERS_CHANGED", "RAID_ROSTER_UPDATE"}, 1, "ScanRaidParty");
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	
	-- Register callbacks
	if QH then
	   QH.RegisterCallback(self, "HealthUpdated", function(event, GUID, newHealth) end)
	end
	
	-- Inject unit frame right-click menu option
	UnitPopupButtons["SHOW_DEATH_REPORTS"] = {
		text = L["Show Acheron Death Reports"],
		dist = 0,
		func = self.ShowDeathReportsFromUnitMenu
	}
	
	for i = 1, #menuTypes do
		tinsert(UnitPopupMenus[menuTypes[i]], #UnitPopupMenus[menuTypes[i]]-1, "SHOW_DEATH_REPORTS")
	end

	self:SecureHook("UnitPopup_ShowMenu")

end


--[[ ---------------------------------------------------------------------------
	 Ace3 disable addon
----------------------------------------------------------------------------- ]]
function Acheron:OnDisable()

	self:DoDisable()

end


--[[ ---------------------------------------------------------------------------
	 Disables addon
----------------------------------------------------------------------------- ]]
function Acheron:DoDisable()

	if self.debug then self:debug("Addon disabled") end

	-- Unregister events
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:UnregisterBucket(self.updateUnitsBucket)

	-- Unregister callbacks
	if QH then
	   QH.UnregisterCallback(self, "HealthUpdated")
	end
	
	-- Remove unit frame right-click menu option
	for j = 1, #menuTypes do
		local t = menuTypes[j]
		for i = 1, #UnitPopupMenus[t] do
			if UnitPopupMenus[t][i] == "SHOW_DEATH_REPORTS" then
				tremove(UnitPopupMenus[t], i)
				break
			end
		end
	end

	self:UnhookAll()
	
end


--[[ ---------------------------------------------------------------------------
	 Check to see if we need to disable for PvP zones
----------------------------------------------------------------------------- ]]
function Acheron:ZONE_CHANGED_NEW_AREA()

	if self:GetProfileParam("disablepvp") then 

		local instanceType = select(2, IsInInstance())
	
		if instanceType == "none" and select(1,GetZonePVPInfo()) == "combat" then
			instanceType = "wintergrasp"
		end
		
		local isPvPZone = (instanceType == "pvp" or instanceType == "arena" or instanceType == "wintergrasp")

		if self:GetProfileParam("enable") and isPvPZone then
			self:SetProfileParam("enable", false)
			self:DoDisable()
		elseif not self:GetProfileParam("enable") and not isPvPZone then
			self:SetProfileParam("enable", true)
			self:DoEnable()
		end
		
	end
	
	if self:GetProfileParam("enable") then 
		self:ScanRaidParty()
	end

end


--[[ ---------------------------------------------------------------------------
	 Hook function for injecting unit frame right-click menu option
----------------------------------------------------------------------------- ]]
function Acheron:UnitPopup_ShowMenu(dropdownMenu, which, unit, name, userData, ...)
	for i=1, UIDROPDOWNMENU_MAXBUTTONS do
		button = _G["DropDownList"..UIDROPDOWNMENU_MENU_LEVEL.."Button"..i];
		if button.value == "SHOW_DEATH_REPORTS" then
		    button.func = UnitPopupButtons["SHOW_DEATH_REPORTS"].func
		end
	end
end


--[[ ---------------------------------------------------------------------------
	 Iterate through party/raid to see what we need to track
----------------------------------------------------------------------------- ]]
function Acheron:ScanRaidParty()

	if self.debug then self:debug("Scanning raid/party") end
	
	local groupType = "party"
	local groupSize = GetNumPartyMembers()

	if GetNumRaidMembers() > 0 then
		groupType = "raid"
		groupSize = 40
	end
	
	local i = 0
	local unit = "player"
		
	while i <= groupSize do
	
		if UnitExists(unit) then
			
			local id = UnitGUID(unit)
			
			if not self.combatLogs[id] then
				self.guidNames[id] = UnitName(unit)
				self.combatLogs[id] = {first = 1, last = 1, logs = {}}
				self.combatLogs[id].logs[1] = {first = 1, last = 0, log = {}}
			end
			
		end
	
		i = i + 1
		unit = groupType..i
	
	end
	
	-- change grouping status
	if (GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) and self.solo == "yes" and self:GetProfileParam("clearongroup") then
		self.solo = "no"
		if Acheron:GetProfileParam("confirmclear") then
			StaticPopup_Show("ACHERON_CLEAR_ALL")
		else 
			Acheron:ClearReports()
		end
	end
	
	if GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0 then
		self.solo = "no"
	else
		self.solo = "yes"
	end
	
end


--[[ ---------------------------------------------------------------------------
	 Handle combat log events
----------------------------------------------------------------------------- ]]
function Acheron:COMBAT_LOG_EVENT_UNFILTERED(event, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
  
	-- Stop if this is a combat log event we don't care about
	if not dstGUID then return end						  
	if not self.combatLogs[dstGUID] then return end
	
	if eventType == "UNIT_DIED" and not UnitIsFeignDeath(self.guidNames[dstGUID]) then
		self:HandleDeathEvent(timeStamp, dstGUID)
	elseif eventType == "SPELL_AURA_APPLIED" or
		   eventType == "SPELL_AURA_REMOVED" or
		   eventType == "SPELL_AURA_APPLIED_DOSE" or
		   eventType == "SPELL_AURA_REMOVED_DOSE"
	then
		self:HandleBuffEvent(event, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	else
		self:HandleHealthEvent(event, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	end

end


--[[ ---------------------------------------------------------------------------
	 Handle death
----------------------------------------------------------------------------- ]]
function Acheron:HandleDeathEvent(timeStamp, dstGUID)

	local latestLog = self.combatLogs[dstGUID].logs[self.combatLogs[dstGUID].last]

	-- If nothing happened before the death, forget it (Spirit of Redemption)
	if latestLog.last == 0 then return end

	-- Make the entry
	self:TrackEvent("DEATH", timeStamp, nil, dstGUID, 0, L["Death"])
	
	-- Start a new log
	self.combatLogTimeline.last = self.combatLogTimeline.last + 1
	self.combatLogTimeline.logs[self.combatLogTimeline.last] = {id = dstGUID, reportNum = self.combatLogs[dstGUID].last, timeStamp = timeStamp}
	
	self.combatLogs[dstGUID].last = self.combatLogs[dstGUID].last + 1
	self.combatLogs[dstGUID].logs[self.combatLogs[dstGUID].last] = {first = 1, last = 0, log = {}}
		
	-- Remove older logs if threshold is set
	local timelineNum = self.combatLogTimeline.last - self.combatLogTimeline.first + 1
	if self:GetProfileParam("numreports") > 0 and timelineNum > self:GetProfileParam("numreports") then
	
		local firstTimeline = self.combatLogTimeline.logs[self.combatLogTimeline.first]
		
		self.combatLogs[firstTimeline.id].logs[firstTimeline.reportNum] = nil
		self.combatLogs[firstTimeline.id].first = firstTimeline.reportNum + 1
		
		self.combatLogTimeline.logs[self.combatLogTimeline.first] = nil
		self.combatLogTimeline.first = self.combatLogTimeline.first + 1
		
	end

end


--[[ ---------------------------------------------------------------------------
	 Handle combat events that affect buffs (damage or healing)
----------------------------------------------------------------------------- ]]
function Acheron:HandleBuffEvent(event, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)

	local spellId = select(1, ...)
	local spell = select(2, ...)
	local auraType = select(4,...)
	local debuffCount = select(5,...) or 1

	if not spell then return end
	
	if self.trace then self:trace("%s: %s x %d", eventType, spell, debuffCount) end
	
	-- special handler for DK "Shadow of Death" ghouling and Priest "Spirit of Redemption"
	if eventType == "SPELL_AURA_APPLIED" and (spellId == 54223 or spellId == 27827) then
		self:HandleDeathEvent(timeStamp, dstGUID)
		return
	elseif eventType == "SPELL_AURA_REMOVED" and spellId == 27827 then
		return
	end
	
	if ((self:GetProfileParam("enablewhitelist") and not self.db.profile.aurawhitelist[spell]) or
		(self:GetProfileParam("enableblacklist") and self.db.profile.aurablacklist[spell]))
	then
		return
	end
	
	local action = spell
	
	if eventType == "SPELL_AURA_REMOVED" then
		action = "-" .. action
	else
		action = "+" .. action
	end
	
	if debuffCount > 1 then
		action = action .. " ("..debuffCount..")"
	end

	local msg = CombatLog_OnEvent(Blizzard_CombatLog_CurrentSettings, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	
	self:TrackEvent(auraType, timeStamp, msg, dstGUID, 0, action)
	
end

--[[ ---------------------------------------------------------------------------
	 Handle combat events that affect health (damage or healing)
----------------------------------------------------------------------------- ]]
function Acheron:HandleHealthEvent(event, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
 	
	-- Initialize values		  
	local amount = nil
	local action = nil
	local missType = nil
	local spellId = select(1, ...)
	local spell = select(2, ...)
	local source = srcName
	local isCrit = false
	local isCrush = false
	local trackType = "DAMAGE"
	
	-- special handling for Priest "Spirit of Redemption"
	if spellId == 27827 then
		return
	end

	-- Extract the useful data	
	local prefix, suffix, special = strsplit("_", eventType)
		
	if prefix == "SPELL" then
		if suffix == "HEAL" or special == "HEAL" then
			amount = select(4, ...)
			isCrit = select(6, ...)
		elseif suffix == "DAMAGE" or special == "DAMAGE" then
			amount = 0 - select(4, ...)
			isCrit = select(10, ...)
		elseif suffix == "MISSED" or special == "MISSED" then
			missType =  select(4, ...)
		end
	elseif suffix == "DAMAGE" then
		if prefix == "SWING" then
			amount = 0 - select(1, ...)
			spell = L["Melee"]
			isCrit = select(7, ...)
			isCrush = select(9, ...)
		elseif prefix == "RANGE" then
			amount = 0 - select(4, ...)
			isCrit = select(10, ...)
		elseif prefix == "ENVIRONMENTAL" then
			amount = 0 - select(2, ...)
			spell = select(1, ...)
			source = L["Environment"]
		end
	elseif prefix == "DAMAGE" and special ~= "MISSED" then
		amount = 0 - select(4, ...)
	elseif suffix == "MISSED" then
		if prefix == "RANGE" then
			missType =  select(4, ...)
		else
			missType =  select(1, ...)
		end
	end
	
	-- Convert miss type to string
	if missType then
		if missType == "DODGE" then amount, action = 0, L["Dodge"] end
		if missType == "PARRY" then amount, action = 0, L["Parry"] end
		if missType == "MISS" then amount, action = 0, L["Miss"] end
		if missType == "RESIST" then amount, action = 0, L["Resist"] end
	end
	
	if not amount and not action then return end
	if amount > 0 then trackType = "HEAL" end
	
	local msg = CombatLog_OnEvent(Blizzard_CombatLog_CurrentSettings, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	
	-- Track the event
	self:TrackEvent(trackType, timeStamp, msg, dstGUID, amount, action, source, spell, isCrit, isCrush)
	
end


--[[ ---------------------------------------------------------------------------
	 Track a combat log event
----------------------------------------------------------------------------- ]]
function Acheron:TrackEvent(eventType, timeStamp, msg, dstGUID, amount, action, source, spell, isCrit, isCrush)

	local dstName = self.guidNames[dstGUID]
	local latestLog = self.combatLogs[dstGUID].logs[self.combatLogs[dstGUID].last]
								
	-- Compose the entry for this event
	local sourceClass = nil
	if source and UnitIsPlayer(source) then
		_, sourceClass = UnitClass(source)
	end
	
	if not amount then amount = 0 end
	if not source then source = spell end
	local curHealth
	if QH then
	   curHealth = QH:UnitHealth(dstName)
	else
	   curHealth = UnitHealth(dstName)
	end
	local maxHealth = UnitHealthMax(dstName)
	local pctHealth = ((curHealth/maxHealth) * 100)
	if pctHealth < 0 then pctHealth = "--%" else pctHealth = format("%3d%%", pctHealth) end
	
	local entry = {
		eventType = eventType,
		timeStamp = timeStamp,
		msg = msg,
		curHealth = curHealth,
		maxHealth = maxHealth,
		pctHealth = pctHealth,
		amount = amount,
		action = action,
		source = source,
		sourceClass = sourceClass,
		spell = spell,
		isCrit = isCrit,
		isCrush = isCrush,
	}
		
	-- Add the entry to the latest log
	latestLog.last = latestLog.last + 1
	latestLog.log[latestLog.last] = entry
	
	-- Remove any entries that are older than our threshold
	for i = latestLog.first, latestLog.last do
	
		if (latestLog.log[i].timeStamp >= (timeStamp - self:GetProfileParam("history"))) then
			latestLog.first = i;
			break
		else
			latestLog.log[i] = nil
		end
	
	end

end


--[[ ---------------------------------------------------------------------------
	 Clear all reports
----------------------------------------------------------------------------- ]]
function Acheron:ClearReports(name)

	if name then
	
		local id = self:GetGUID(name)
		self.combatLogs[id] = {first = 1, last = 1, logs = {}}
		self.combatLogs[id].logs[1] = {first = 1, last = 0, log = {}}
		
		local newTimeline = {first = 1, last = 0, logs = {}}
		
		for i = self.combatLogTimeline.first, self.combatLogTimeline.last do
			if self.guidNames[self.combatLogTimeline.logs[i].id] ~= name then
				newTimeline.last = newTimeline.last + 1
				newTimeline.logs[newTimeline.last] = self.combatLogTimeline.logs[i]
			end
		end
		
		self.combatLogTimeline = newTimeline
		
	else
	
		self.combatLogs = {}
		self.combatLogTimeline = {first = 1, last = 0, logs = {}}
		
	end
	
	self:ScanRaidParty()
	
	if self.frame:IsVisible() then
		self:ShowDeathReports()
	end

end


--[[ ---------------------------------------------------------------------------
	 Get GUID from name
----------------------------------------------------------------------------- ]]
function Acheron:GetGUID(name)

	local id = UnitGUID(name)
	
	if name and not id then
		for logid,logname in pairs(Acheron.guidNames) do
			if logname == name then id = logid end
		end
	end
	
	return id

end
